import jax
import jax.numpy as jnp

from .mhrm import fit_mhrm
from .utils import cal_p12


class Base():
    def __init__(self,
                 data,
                 n_factors,
                 patterns=None,
                 freq=None,
                 init_frac=None,
                 verbose=None,
                 key=None):
        if not isinstance(data, type(jnp.array)):
            data = jnp.array(data)
        n_cases = data.shape[0]
        n_items = data.shape[1]
        ns_cats = list(jnp.nanmax(data, axis=0).astype("int") + 1)
        max_cats = max(ns_cats)
        dtype = data.dtype
        y = jax.nn.one_hot(
            x=data,
            num_classes=max_cats,
            dtype=dtype)
        del data
        if min(ns_cats) == max_cats:
            cats = "equal"
        else:
            cats = "unequal"
        if isinstance(patterns, type(None)):
            analysis = "exploratory"
            patterns = {"loading": None, "corr": None}
        else:
            analysis = "confirmatory"
            if "loading" not in patterns:
                patterns["loading"] = None
            if "corr" not in patterns:
                patterns["corr"] = None
        if isinstance(freq, type(None)):
            freq = jnp.full(
                shape=(n_cases,),
                fill_value=1.,
                dtype=dtype)
        else:
            if not isinstance(freq, type(jnp.array)):
                freq = jnp.array(
                    freq, dtype=dtype)
        if isinstance(verbose, type(None)):
            verbose = True
        if isinstance(key, type(None)):
            key = jax.random.PRNGKey(0)
        if isinstance(init_frac, type(None)):
            p1, p2 = cal_p12(y, freq)
        else:
            key, subkey = jax.random.split(key)
            init_idx = jax.random.choice(
                key=subkey,
                a=n_cases,
                shape=(int(n_cases * init_frac),),
                replace=False)
            p1, p2 = cal_p12(
                y[init_idx, ...],
                freq[init_idx, ...])
        stats = {"p1": p1, "p2": p2}
        info = {"n_cases": n_cases,
                "n_items": n_items,
                "n_factors": n_factors,
                "ns_cats": ns_cats,
                "max_cats": max_cats,
                "analysis": analysis,
                "patterns": patterns,
                "cats": cats,
                "dtype": dtype}
        self.y, self.freq = y, freq
        self.info = info
        self.stats = stats
        self.key, self.verbose = key, verbose

    def print_init(self):
        if self.verbose:
            print("A", self.__class__.__name__,
                  "Object is Initialized for",
                  self.info["analysis"].capitalize(),
                  "Analysis.")
            print(" + Number of Cases: %.0f" % (self.info["n_cases"]))
            print(" + Number of Items: %.0f" % (self.info["n_items"]))
            print(" + Number of Factors: %.0f" % (self.info["n_factors"]))
            if self.info["cats"] == "equal":
                print(" + Number of Categories: %.0f" % (self.info["max_cats"]))
            else:
                print(" + Number of Categories: %.0f-%.0f" % (min(self.info["ns_cats"]), self.info["max_cats"]))


class Ordinal(Base):
    def init_masks(self):
        self.masks = {}
        if self.info["analysis"] == "exploratory":
            self.masks["loading"] = jnp.ones(
                (self.info["n_items"],
                 self.info["n_factors"]),
                dtype=self.info["dtype"])
        else:
            if isinstance(
                    self.info["patterns"]["loading"], type(None)):
                self.masks["loading"] = jnp.ones(
                    (self.info["n_items"],
                     self.info["n_factors"]),
                    dtype=self.info["dtype"])
            else:
                row_idx = []
                col_idx = []
                for key, values in self.info["patterns"]["loading"].items():
                    for value in values:
                        row_idx.append(value)
                        col_idx.append(key)
                self.masks["loading"] = jax.ops.index_update(
                    jnp.zeros(
                        (self.info["n_items"],
                         self.info["n_factors"]),
                        dtype=self.info["dtype"]),
                    (row_idx, col_idx), 1.)
        if self.info["analysis"] == "exploratory":
            self.masks["corr"] = jnp.zeros(
                (self.info["n_factors"],
                 self.info["n_factors"]),
                dtype=self.info["dtype"])
        else:
            if isinstance(
                    self.info["patterns"]["corr"], type(None)):
                self.masks["corr"] = jnp.ones(
                    (self.info["n_factors"],
                     self.info["n_factors"]),
                    dtype=self.info["dtype"]) - jnp.eye(
                    self.info["n_factors"],
                    dtype=self.info["dtype"])
            else:
                row_idx = []
                col_idx = []
                for key, values in self.info["patterns"]["corr"].items():
                    for value in values:
                        if (key != value):
                            row_idx.append(value)
                            col_idx.append(key)
                            col_idx.append(value)
                            row_idx.append(key)
                self.masks["corr"] = jax.ops.index_update(
                    jnp.zeros(
                        (self.info["n_factors"],
                         self.info["n_factors"]),
                        dtype=self.info["dtype"]),
                    (row_idx, col_idx), 1.)

    def init_params(self):
        self.params = {}

        def init_loading(p1, p2, n_factors):
            max_cats = p1.shape[1]
            k = jnp.arange(max_cats)
            m1 = jnp.sum(p1 * k, axis=1)
            m2 = jnp.sum(p2 * jnp.outer(k, k)[None, :, None, :], axis=(1, 3))
            s = m2 - jnp.outer(m1, m1)
            d = jnp.sqrt(jnp.diagonal(s))
            r = s * jnp.outer(d, d)
            eigval, eigvec = jnp.linalg.eigh(r)
            loading = eigvec[:, -n_factors:]
            return loading

        if self.info["analysis"] == "exploratory":
            self.params["loading"] = init_loading(
                self.stats["p1"],
                self.stats["p2"],
                self.info["n_factors"])
        else:
            self.params["loading"] = jax.random.uniform(
                self.key, self.masks["loading"].shape) * self.masks["loading"]
        self.params["corr"] = jnp.eye(
            self.info["n_factors"],
            dtype=self.info["dtype"])

    def init_eta(self):
        eta = self.y.argmax(axis=-1) @ self.params["loading"]
        eta = (eta - eta.mean(axis=0)) / eta.std(axis=0)
        self.eta = eta

    def fit(self,
            lr=None,
            max_iters=None,
            stem_iters=None,
            tol=None,
            window_size=None,
            n_chains=None,
            n_warmups=None,
            jump_std=None,
            jump_change=None,
            target_rate=None,
            gain_decay=None,
            corr_update=None,
            batch_size=None,
            cycling=None,
            verbose=None,
            key=None,
            params=None,
            masks=None):
        if isinstance(lr, type(None)):
            lr = 1.
        if isinstance(max_iters, type(None)):
            max_iters = 500
        if isinstance(stem_iters, type(None)):
            stem_iters = 200
        if isinstance(tol, type(None)):
            tol = 10 ** (-4)
        if isinstance(window_size, type(None)):
            window_size = 3
        if isinstance(n_chains, type(None)):
            n_chains = 1
        if isinstance(n_warmups, type(None)):
            n_warmups = 5
        if isinstance(jump_std, type(None)):
            jump_std = 2.4 / jnp.sqrt(self.info["n_factors"])
        if isinstance(jump_change, type(None)):
            jump_change = .01
        if isinstance(target_rate, type(None)):
            target_rate = .23
        if isinstance(gain_decay, type(None)):
            gain_decay = 1.0
        if isinstance(corr_update, type(None)):
            corr_update = "gd"
        if isinstance(verbose, type(None)):
            verbose = self.verbose
        if isinstance(key, type(None)):
            key = self.key
        if isinstance(params, type(None)):
            params = self.params
        if isinstance(masks, type(None)):
            masks = self.masks
        y, freq = self.y, self.freq
        eta = self.eta
        crf = self.crf
        params, aparams, eta, trace = fit_mhrm(
            lr=lr,
            max_iters=max_iters,
            stem_iters=stem_iters,
            tol=tol,
            window_size=window_size,
            n_chains=n_chains,
            n_warmups=n_warmups,
            jump_std=jump_std,
            jump_change=jump_change,
            target_rate=target_rate,
            gain_decay=gain_decay,
            corr_update=corr_update,
            batch_size=batch_size,
            cycling=cycling,
            verbose=verbose,
            key=key,
            params=params,
            masks=masks,
            y=y,
            eta=eta,
            freq=freq,
            crf=crf)
        self.params = params
        self.aparams = aparams
        self.eta = eta
        self.trace = trace
        if verbose:
            if self.trace["is_nan"]:
                print("`NaN` Occurs after %.0f Iterations (%.2f sec)." % (
                    self.trace["n_iters"], self.trace["fit_time"]))
                print("Possible Solutions Include:")
                print("+ Try Other `key`.")
                print("+ Try Smaller `lr`.")
                print("+ Set `corr_update='empirical'`.")
            else:
                if self.trace["is_converged"]:
                    print("Converged after %.0f Iterations (%.2f sec)." % (
                        self.trace["n_iters"], self.trace["fit_time"]))
                else:
                    print("Not Converged after %.0f Iterations (%.2f sec)." % (
                        self.trace["n_iters"], self.trace["fit_time"]))
                    print("Possible Solutions Include:")
                    print("+ Try Larger `max_iters`.")
                    print("+ Try Larger `n_chains`.")
        return self

    def transform(
            self,
            data=None,
            n_chains=None,
            n_warmups=None,
            jump_std=None,
            batch_size=None,
            verbose=None,
            key=None):
        params = self.params
        crf = self.crf
        if isinstance(data, type(None)):
            y, freq, eta = self.y, self.freq, self.eta
            n_cases = y.shape[0]
        else:
            if not isinstance(data, type(jnp.array)):
                data = jnp.array(data)
            y = jax.nn.one_hot(
                x=data,
                num_classes=self.info["max_cats"],
                dtype=self.info["dtype"])
            n_cases = y.shape[0]
            del data
            freq = jnp.full(
                shape=(n_cases,),
                fill_value=1.,
                dtype=self.info["dtype"])
            eta = y.argmax(axis=-1) @ params["loading"]
            eta = (eta - eta.mean(axis=0)) / eta.std(axis=0)
        if isinstance(n_chains, type(None)):
            n_chains = 100
        if isinstance(n_warmups, type(None)):
            n_warmups = 100
        if isinstance(jump_std, type(None)):
            jump_std = self.trace["jump_std"]
        if isinstance(verbose, type(None)):
            verbose = self.verbose
        if isinstance(key, type(None)):
            key = self.key
        key, subkey = jax.random.split(key)
        eta3d = jnp.repeat(eta[None, ...], n_chains, axis=0)
        if isinstance(batch_size, type(None)):
            eta3d, accept_rate = conduct_mcmc(
                subkey, n_warmups, jump_std,
                eta3d, y, freq, params, crf)
        else:
            n_batches = int(jnp.ceil(n_cases / batch_size))
            batch_slices = [
                slice(batch_size * i, min(batch_size * (i + 1), n_cases), 1) for i in range(n_batches)]
            sum_freq = freq.sum()
            accept_rate = jnp.zeros(())
            key, subkey = jax.random.split(key)
            for batch_slice in batch_slices:
                y_batch, eta3d_batch, freq_batch = y[batch_slice, ...], eta3d[:, batch_slice, :], freq[batch_slice]
                key, subkey = jax.random.split(key)
                eta3d_batch, accept_rate_batch = conduct_mcmc(
                    subkey, n_warmups, jump_std,
                    eta3d_batch, y_batch, freq_batch, params, crf)
                accept_rate = accept_rate + (freq_batch.sum() / sum_freq) * accept_rate_batch
                eta3d = jax.ops.index_update(
                    eta3d, jax.ops.index[:, batch_slice, :], eta3d_batch)
        eta = jnp.mean(eta3d, axis=0)
        if verbose:
            print("Data are Transformed to Factor Scores by EAP.")
            print("+ Number of Cases: %.0f" % (n_cases))
            print("+ Number of Chains: %.0f" % (n_chains))
            print("+ Number of Warm-Up: %.0f" % (n_warmups))
            print("+ Accept Rate: %.3f" % (accept_rate))
        return eta


import jax
import jax.numpy as jnp
from jax import jit, vmap


@jit
def cal_multinomial_logpmf(k, p):
    multinomial_logpmf = jax.scipy.special.xlogy(k, p).sum(axis=-1)
    return multinomial_logpmf


@jit
def cal_normal_logpdf(x, mean, cov):
    l = jax.lax.linalg.cholesky(jnp.linalg.inv(cov))
    z = (x - mean) @ l
    normal_logpdf = -0.5 * (z.shape[-1] * jnp.log(2 * jnp.pi) + jnp.square(z).sum(-1)) + jnp.sum(jnp.log(l.diagonal()))
    return normal_logpdf


@jit
def cal_p12(y, freq):
    y_freq = y * freq[..., None, None]
    f1 = jnp.sum(y_freq, axis=0) + 1.
    p1 = f1 / jnp.sum(f1, axis=1)[..., None]
    f2 = vmap(lambda x_i: y_freq.T @ x_i, in_axes=(1,))(y)
    p2 = f2 / jnp.sum(f2, axis=(1, 3))[:, None, :, None]
    return p1, p2



import time
from functools import partial

import jax
import jax.numpy as jnp
from jax import grad, jit, vmap
from tqdm.auto import tqdm

from .utils import cal_multinomial_logpmf



@partial(jit, static_argnums=(4,))
def cal_closs2d_i(
        params, y, eta2d, freq, crf):
    closs_i = cal_floss2d_i(params, y, eta2d, freq, crf) + cal_bloss2d_i(params, eta2d, freq)
    return closs_i


@partial(jit, static_argnums=(4,))
def cal_closs2d(
        params, y, eta2d, freq, crf):
    closs = cal_closs2d_i(params, y, eta2d, freq, crf).sum() / freq.sum()
    return closs


@partial(jit, static_argnums=(4,))
def cal_dcloss2d(
        params, y, eta2d, freq, crf):
    dcloss = grad(cal_closs2d, argnums=0)(params, y, eta2d, freq, crf)
    return dcloss


@partial(jit, static_argnums=(4,))
def cal_dcloss2d_i(
        params, y, eta2d, freq, crf):
    dcloss_i = jax.jacfwd(cal_closs2d_i, argnums=0)(params, y, eta2d, freq, crf)
    return dcloss_i


@partial(jit, static_argnums=(4,))
def cal_d2closs2d(
        params, y, eta2d, freq, crf):
    return jax.jacfwd(cal_dcloss2d, argnums=0)(params, y, eta2d, freq, crf)


@partial(jit, static_argnums=(4,))
def cal_closs3d_i(
        params, y, eta3d, freq, crf):
    closs_i = vmap(cal_closs2d_i,
                   in_axes=(None, None, 0, None, None))(
        params, y, eta3d, freq, crf)
    return closs_i


@partial(jit, static_argnums=(4,))
def cal_closs3d(
        params, y, eta3d, freq, crf):
    closs = cal_closs3d_i(params, y, eta3d, freq, crf).mean(axis=0).sum() / freq.sum()
    return closs


@partial(jit, static_argnums=(4,))
def cal_dcloss3d(
        params, y, eta3d, freq, crf):
    dcloss = grad(cal_closs3d, argnums=0)(params, y, eta3d, freq, crf)
    return dcloss


@partial(jit, static_argnums=(4,))
def cal_dcloss3d_i(
        params, y, eta3d, freq, crf):
    dcloss_i = vmap(cal_dcloss2d_i,
                    in_axes=(None, None, 0, None, None))(
        params, y, eta3d, freq, crf)
    return dcloss_i


@partial(jit, static_argnums=(4,))
def cal_d2closs3d(
        params, y, eta3d, freq, crf):
    d2closs = jax.jacfwd(cal_dcloss3d, argnums=0)(params, y, eta3d, freq, crf)
    return d2closs


@partial(jit, static_argnums=(4,))
def cal_floss2d_i(
        params, y, eta2d, freq, crf):
    floss_i = -jnp.sum(
        cal_multinomial_logpmf(
            k=y,
            p=crf(eta2d, params)),
        axis=-1) * freq
    return floss_i


@jit
def cal_bloss2d_i(
        params, eta2d, freq):
    n_factors = eta2d.shape[-1]
    bloss_i = - jax.scipy.stats.multivariate_normal.logpdf(
        x=eta2d,
        mean=jnp.zeros(n_factors),
        cov=params["corr"]) * freq
    return bloss_i


@jit
def cal_bloss3d_i(
        params, eta3d, freq):
    bloss_i = vmap(cal_bloss2d_i,
                   in_axes=(None, 0, None))(
        params, eta3d, freq)
    return bloss_i


@jit
def cal_bloss3d(
        params, eta3d, freq):
    bloss = cal_bloss3d_i(params, eta3d, freq).mean(axis=0).sum() / freq.sum()
    return bloss


def adjust_jump_std(
        jump_std, accept_rate, target_rate, jump_change):
    if accept_rate > (target_rate + .01):
        jump_std = jump_std + jump_change
    elif accept_rate < (target_rate - .01):
        jump_std = jump_std - jump_change
    else:
        pass
    return jump_std


def set_timers(
        max_iters, stem_iters):
    timer1 = tqdm(
        total=stem_iters,
        disable=False,
        position=0)
    timer2 = tqdm(
        total=max_iters - stem_iters,
        disable=False,
        position=1)
    return timer1, timer2


def update_timers(
        timer1, timer2, n_iter, stem_iters, trace):
    if n_iter <= stem_iters:
        timer1.set_postfix(
            {"Accept Rate": jnp.round(trace["accept_rate"][-1], 2),
             'Loss': jnp.round(trace["closs"][-1], 2)})
        timer1.set_description("Stage 1")
        timer1.update(1)
    else:
        timer2.set_postfix(
            {"Accept Rate": jnp.round(trace["accept_rate"][-1], 2),
             'Loss': jnp.round(trace["closs"][-1], 2)})
        timer2.set_description("Stage 2")
        timer2.update(1)


@partial(jit, static_argnums=(7,))
def conduct_mcmc(
        key, n_warmups, jump_std,
        eta3d, y, freq, params, crf):
    accept_rate = 0
    n_factors = eta3d.shape[-1]

    def sample_eta3d(i, value):
        key, eta3d, accept_rate = value
        key, subkey = jax.random.split(key)
        eta3d_new = jax.random.multivariate_normal(
            key=subkey,
            mean=eta3d,
            cov=jnp.diag(jnp.repeat(jump_std**2, n_factors)),
            shape=eta3d.shape[:-1])
        ratio = jnp.exp(
            - cal_closs3d_i(params, y, eta3d_new, freq, crf) + cal_closs3d_i(params, y, eta3d, freq, crf))
        key, subkey = jax.random.split(key)
        accept = jax.random.bernoulli(
            key=subkey,
            p=jnp.minimum(ratio, 1))
        eta3d = jnp.where(
            jnp.repeat(
                accept[..., None],
                n_factors,
                axis=-1),
            x=eta3d_new,
            y=eta3d)
        accept_rate = accept.mean()
        return key, eta3d, accept_rate

    key, eta3d, accept_rate = jax.lax.fori_loop(
        0, n_warmups + 1, sample_eta3d, (key, eta3d, accept_rate))
    return eta3d, accept_rate


@jit
def cal_cov_eta3d(eta3d):
    cov_eta = jnp.mean(
        vmap(lambda eta:
             eta.T @ eta / eta.shape[0], in_axes=0)(eta3d),
        axis=0)
    return cov_eta


@jit
def ls_corr(
        lr, gain, params, dparams, masks, eta3d, freq):
    c_armijo = 0.001
    d_armijo = 1.0
    lr_armijo = 0
    iter_armijo = 0
    corr = params["corr"]
    bloss = cal_bloss3d(params, eta3d, freq)
    value = (d_armijo, params, lr_armijo, iter_armijo)

    def body_fun(value):
        d_armijo, params, lr_armijo, iter_armijo = value
        lr_armijo = (0.5 ** iter_armijo) * lr
        params["corr"] = corr - lr_armijo * gain * dparams["corr"] * masks["corr"]
        d_armijo = cal_bloss3d(params, eta3d, freq) - bloss + c_armijo * lr_armijo * jnp.sum(dparams["corr"] ** 2)
        iter_armijo = iter_armijo + 1
        return d_armijo, params, lr_armijo, iter_armijo

    def cond_fun(value):
        d_armijo, params, lr_armijo, iter_armijo = value
        cond = d_armijo > 0.
        return cond

    d_armijo, params, lr_armijo, iter_armijo = jax.lax.while_loop(cond_fun, body_fun, value)
    return params["corr"], lr_armijo


@jit
def project_corr(
        params, masks):
    scale = jnp.sqrt(params["corr"].diagonal())
    params["corr"] = params["corr"] / jnp.outer(scale, scale)
    params["corr"] = params["corr"] * masks["corr"] + jnp.diag(params["corr"].diagonal())
    return params


def update_params(
        lr, gain,
        stage, corr_update,
        params, dparams, masks,
        eta3d, freq):
    params["intercept"] = params["intercept"] - lr * gain * dparams["intercept"] * masks["intercept"]
    params["loading"] = params["loading"] - lr * gain * dparams["loading"] * masks["loading"]
    if jnp.sum(masks["corr"]) >= 1.0:
        if stage == 1:
            params["corr"] = cal_cov_eta3d(eta3d)
            params = project_corr(params, masks)
        else:
            if corr_update == "empirical":
                corr = params["corr"]
                params["corr"] = cal_cov_eta3d(eta3d)
                params = project_corr(params, masks)
                params["corr"] = ((1 - lr * gain) * corr) + (lr * gain * params["corr"])
            else:
                dparams["corr"] = dparams["corr"] + dparams["corr"].T - jnp.diag(dparams["corr"].diagonal())
                if corr_update == "gd":
                    params["corr"] = params["corr"] - lr * gain * dparams["corr"] * masks["corr"]
                elif corr_update == "gd_ls":
                    params["corr"], lr_armijo = ls_corr(
                        lr, gain, params, dparams, masks, eta3d, freq)
                else:
                    pass
    else:
        pass
    return params


def fit_mhrm(
        lr,
        max_iters,
        stem_iters,
        tol,
        window_size,
        n_chains,
        n_warmups,
        jump_std,
        jump_change,
        target_rate,
        gain_decay,
        corr_update,
        batch_size,
        cycling,
        verbose,
        key,
        params,
        masks,
        y,
        eta,
        freq,
        crf):
    start = time.time()
    eta3d = jnp.repeat(eta[None, ...], 1, axis=0)
    if not isinstance(batch_size, type(None)):
        n_cases = y.shape[0]
        n_batches = int(jnp.ceil(n_cases / batch_size))
        batch_slices = [
            slice(batch_size * i, min(batch_size * (i + 1), n_cases), 1) for i in range(n_batches)]
        whole_idx = jnp.arange(n_cases)
        if isinstance(cycling, type(None)):
            cycling = True
    aparams = {key: jnp.zeros(value.shape) for key, value in params.items()}
    trace = {"accept_rate": [], "closs": [], "change_param": []}
    stage = 1
    gain = 1.
    is_converged = False
    is_nan = False
    if verbose:
        timer1, timer2 = set_timers(max_iters, stem_iters)
    for n_iters in range(1, max_iters + 1):
        if n_iters == stem_iters + 1:
            stage = 2
            eta3d = jnp.repeat(eta3d, n_chains, axis=0)
        if stage == 2:
            sa_count = (n_iters - stem_iters)
            gain = 1. / (sa_count ** gain_decay)
        temp = params.copy()
        if isinstance(batch_size, type(None)):
            key, subkey = jax.random.split(key)
            eta3d, accept_rate = conduct_mcmc(
                subkey, n_warmups, jump_std,
                eta3d, y, freq, params, crf)
            dparams = cal_dcloss3d(
                params, y, eta3d, freq, crf)
            params = update_params(
                lr, gain,
                stage, corr_update,
                params, dparams, masks,
                eta3d, freq)
        else:
            sum_freq = freq.sum()
            accept_rate = jnp.zeros(())
            key, subkey = jax.random.split(key)
            whole_idx = jax.random.permutation(subkey, whole_idx)
            if cycling:
                for batch_slice in batch_slices:
                    batch_idx = whole_idx[batch_slice]
                    y_batch, eta3d_batch, freq_batch = y[batch_idx, ...], eta3d[:, batch_idx, :], freq[batch_idx]
                    key, subkey = jax.random.split(key)
                    eta3d_batch, accept_rate_batch = conduct_mcmc(
                        subkey, n_warmups, jump_std,
                        eta3d_batch, y_batch, freq_batch, params, crf)
                    dparams = cal_dcloss3d(
                        params, y_batch, eta3d_batch, freq_batch, crf)
                    params = update_params(
                        lr, gain,
                        stage, corr_update,
                        params, dparams, masks,
                        eta3d_batch, freq_batch)
                    accept_rate = accept_rate + (freq_batch.sum() / sum_freq) * accept_rate_batch
                    eta3d = jax.ops.index_update(
                        eta3d, jax.ops.index[:, batch_idx, :], eta3d_batch)
            else:
                batch_idx = whole_idx[batch_slices[0]]
                y_batch, eta3d_batch, freq_batch = y[batch_idx, ...], eta3d[:, batch_idx, :], freq[batch_idx]
                key, subkey = jax.random.split(key)
                eta3d_batch, accept_rate_batch = conduct_mcmc(
                    subkey, n_warmups, jump_std,
                    eta3d_batch, y_batch, freq_batch, params, crf)
                dparams = cal_dcloss3d(
                    params, y_batch, eta3d_batch, freq_batch, crf)
                params = update_params(
                    lr, gain,
                    stage, corr_update,
                    params, dparams, masks,
                    eta3d_batch, freq_batch)
                accept_rate = accept_rate_batch
                eta3d = jax.ops.index_update(
                    eta3d, jax.ops.index[:, batch_idx, :], eta3d_batch)
        eta3d = eta3d / jnp.sqrt(params["corr"].diagonal())
        closs = cal_closs3d(params, y, eta3d, freq, crf)
        dparams = {key: params[key] - temp[key] for key in params.keys()}
        change_param = max(
            [jnp.max(
                jnp.abs(
                    dparams[key] * masks[key])) for key in dparams.keys()])
        trace["accept_rate"].append(accept_rate)
        trace["change_param"].append(change_param)
        trace["closs"].append(closs)
        if verbose:
            update_timers(
                timer1, timer2, n_iters, stem_iters, trace)
        if jnp.isnan(closs):
            is_nan = True
            break
        if stage == 1:
            jump_std = adjust_jump_std(
                jump_std, accept_rate, target_rate, jump_change)
        else:
            aparams = {key: ((sa_count - 1.) / sa_count) * aparams[key] + (1. / sa_count) * params[key]
                       for key in aparams.keys()}
            if max(trace["change_param"][-window_size:]) < tol:
                is_converged = True
                break
    end = time.time()
    fit_time = end - start
    eta = jnp.mean(eta3d, axis=0)
    trace["jump_std"] = jump_std
    trace["n_iters"] = n_iters
    trace["is_converged"] = is_converged
    trace["is_nan"] = is_nan
    trace["fit_time"] = fit_time
    return params, aparams, eta, trace
